import re
from collections import OrderedDict, defaultdict, ChainMap
import yaml
from .rec import consume_rec


class Dumper(yaml.Dumper):
    def ignore_aliases(self, data):
        return True


class Loader(yaml.Loader):
    pass


def _represent_odict(dumper, instance):
    return dumper.represent_mapping("tag:yaml.org,2002:map", instance.items())


def _construct_odict(loader, node):
    return OrderedDict(loader.construct_pairs(node))


def _represent_str(dumper, instance, _rx=re.compile("[#:]")):
    style = None
    if "\n" in instance:
        style = "|"
    else:
        m = _rx.search(instance)
        if m is not None:
            style = "'"
    return dumper.represent_scalar("tag:yaml.org,2002:str", instance, style=style)


def load(stream, *, buffered=False):
    return yaml.load_all(stream, Loader=Loader)


def dump(d, fp, *, squash_level=0, raw=False, extra_kwargs=None):
    extra_kwargs = extra_kwargs or {}
    default_flow_style = extra_kwargs.get("indent", None) is None
    allow_unicode = not extra_kwargs.get("ensure_ascii", False)

    def _dump(d):
        yaml.dump(
            d,
            fp,
            Dumper=Dumper,
            allow_unicode=allow_unicode,
            default_flow_style=default_flow_style,
        )

    consume_rec(d, _dump, n=squash_level)


Loader.add_constructor("tag:yaml.org,2002:map", _construct_odict)
for dict_class in [OrderedDict, defaultdict, ChainMap]:
    Dumper.add_representer(dict_class, _represent_odict)
Dumper.add_representer(str, _represent_str)



from jqfpy.helpermodule import HelperModule, Context
from jqfpy.accessor import Accessor


class Getter:
    def __init__(self, d, *, accessor=Accessor()):
        self.d = d
        self.accessor = accessor

    def get(self, k=None, d=None, default=None):
        d = d or self.d
        if k is None:
            return d
        access_keys = self.accessor.split_key(k)
        if access_keys and access_keys[0] == "[]":
            rest_keys = access_keys[1:]
            return [self.accessor.access(rest_keys, x, default) for x in d]
        else:
            return self.accessor.access(access_keys, d, default)

    __call__ = get


def create_pycode(fnname, code):
    lines = ["def {}(get, h=None):".format(fnname)]
    lines.extend([line.strip() for line in code.split(";")])
    lines[-1] = "return {}".format(lines[-1])
    pycode = "\n    ".join(lines)
    return pycode


def exec_pycode(fnname, pycode):
    env = {}
    exec(pycode, env)
    return env[fnname]


def create_context(*, here=None, extra_kwargs=None) -> Context:
    return Context(here=here, dump_extra_kwargs=extra_kwargs)


def transform(ctx, fn, d, *, additionals=None):
    getter = Getter(d)
    h = HelperModule(ctx, getter, additionals=additionals)
    return fn(getter, h=h)



import os
import itertools
from collections import OrderedDict
from .langhelpers import reify
from . import loading
from . import accessor
from . import _tree as tree


class Context:
    def __init__(self, *, here=None, dump_extra_kwargs=None):
        self.dump_extra_kwargs = dump_extra_kwargs
        self.here = here
        self.history = [None]
        if here is not None:
            self.chdir(here)

    def chdir(self, here):
        if self.history[-1] != here:
            self.history.append(here)
            os.chdir(here)

    def get_module(self, rf, format=None):
        return loading.get_module(rf, default_format=format)

    @reify
    def loader(self) -> "Loader":
        return Loader(self)

    @reify
    def dumper(self) -> "Dumper":
        return Dumper(self)


class Dumper:
    def __init__(self, ctx: Context):
        self.ctx = ctx

    def dumpfile(self, filename, data, *, raw=False, here=None):
        here = here or self.ctx.here
        if here is not None:
            self.ctx.chdir(here)  # xxx

        with open(filename, "w") as wf:
            m = self.ctx.get_module(wf, format=format)
            m.dump(data, fp=wf, raw=raw, extra_kwargs=self.ctx.dump_extra_kwargs)


class Loader:
    def __init__(self, ctx: Context):
        self.ctx = ctx

    def loadfile(self, filename: str, *, format="json", here=None):
        here = here or self.ctx.here
        if here is not None:
            self.ctx.chdir(here)  # xxx

        with open(filename) as rf:
            m = self.ctx.get_module(rf, format=format)
            return next(m.load(rf))


class HelperModule:
    def __init__(self, ctx, getter, *, factory=OrderedDict, additionals=None):
        self.getter = getter
        self.accessor = getter.accessor  # xxx
        self.factory = factory
        self.additionals = additionals
        self.ctx = ctx

    def __getattr__(self, k):
        if self.additionals is None:
            raise AttributeError(k)
        return getattr(self.additionals, k)

    @property
    def d(self):
        return self.getter.d

    def dumpfile(self, filename, data, *, raw=False, here=None):
        try:
            return self.ctx.dumper.dumpfile(filename, data, raw=raw, here=here)
        except FileNotFoundError as e:
            raise RuntimeError(
                "{e} (where={where!r})".format(e=e, where=self.ctx.history[-1])
            )

    def loadfile(self, filename, *, format="json", here=None):
        try:
            return self.ctx.loader.loadfile(filename, format=format, here=here)
        except FileNotFoundError as e:
            raise RuntimeError(
                "{e} (where={where!r})".format(e=e, where=self.ctx.history[-1])
            )

    def pick(self, *ks, d=None, default=None):
        d = d or self.d
        return pick(
            d, ks, default=default, factory=self.factory, accessor=self.accessor
        )

    def omit(self, *ks, d=None):
        d = d or self.d
        return omit(d, ks, factory=self.factory, accessor=self.accessor)

    def flatten(self, L, *, n):
        for _ in range(n):
            L = flatten1(L)
        return L

    def flatten1(self, L):
        return flatten1(L)

    def chunk(self, L, *, n):
        return list(chunk(L, n=n))


def _build_dict(triples, *, factory):
    d = factory()
    for access_keys, build_keys, v in triples:
        cursor = d
        if not build_keys:
            build_keys = access_keys

        for k in build_keys[:-1]:
            if k not in cursor:
                cursor[k] = factory()
            cursor = cursor[k]
        cursor[build_keys[-1]] = v
    return d


def pick(d, ks, *, default=None, factory=OrderedDict, accessor=accessor.Accessor()):
    gen = _pick_gen(d, ks, default=default, accessor=accessor)
    return _build_dict(gen, factory=factory)


def _pick_gen(d, ks, *, default, accessor):
    for k in ks:
        access_keys, build_keys = accessor.split_key_pair(k)
        yield access_keys, build_keys, accessor.access(access_keys, d, default=default)


def omit(d, ks, *, factory=OrderedDict, accessor=accessor.Accessor()):
    t = tree.build_tree([accessor.split_key(k) for k in ks])
    gen = _omit_gen(d, t, [])
    return _build_dict(gen, factory=factory)


def _omit_gen(d, t, hist):
    for k in d.keys():
        if k in t:
            hist.append(k)
            yield from _omit_gen(d[k], t.children[k], hist=hist)
            hist.pop()
        elif k in t.leafs:
            continue
        else:
            hist.append(k)
            yield hist[:], [], d[k]
            hist.pop()


def flatten1(L):
    return list(itertools.chain.from_iterable(L))


def chunk(iterable, n):
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield tuple(itertools.chain((first_el,), chunk_it))



from collections import OrderedDict


class Node:
    def __init__(self):
        self.children = OrderedDict()
        self.leafs = set()

    def __contains__(self, k):
        return k in self.children

    def subtree(self, k):
        if k not in self:
            self.children[k] = self.__class__()
        return self.children[k]

    def add_leaf(self, v):
        self.leafs.add(v)


def build_tree(L):
    t = Node()

    for ks in L:
        cursor = t
        for k in ks[:-1]:
            cursor = cursor.subtree(k)
        cursor.add_leaf(ks[-1])
    return t


def dump_tree(t, indent=""):
    for k, st in t.children.items():
        print("{}{}".format(indent, k))
        dump_tree(st, indent=indent + "  ")
    for v in t.leafs:
        print("{}{}".format(indent, v))



# todo: strict version
class Accessor:
    def split_key(self, k, *, sep="/"):
        return [normalize_json_pointer(x) for x in k.lstrip(sep).split(sep)]

    def split_key_pair(self, k, *, sep="@"):
        if sep not in k:
            return self.split_key(k), []
        else:
            access_keys, build_keys = k.split(sep, 1)
            return self.split_key(access_keys), self.split_key(build_keys)

    def access(self, access_keys, d, default=None):
        for i, k in enumerate(access_keys):
            if k == "*":
                if len(access_keys) - 1 == i:
                    continue  # on last, no effect
                else:
                    next_key = access_keys[i + 1]
                    rest_keys = access_keys[i + 1 :]
                    for gk, v in d.items():
                        if hasattr(v, "__contains__") and next_key in v:
                            return self.access(rest_keys, d[gk])
                    return default
            elif k == "*[]":
                if len(access_keys) - 1 == i:
                    continue  # on last, no effect
                else:
                    next_key = access_keys[i + 1]
                    rest_keys = access_keys[i + 1 :]
                    candidates = []
                    for gk, v in d.items():
                        if hasattr(v, "__contains__") and next_key in v:
                            candidates.append(v)
                    if candidates:
                        return [self.access(rest_keys, v) for v in candidates]
                    return default
            elif k.endswith("[]"):
                k = k.rstrip("[]")
                rest_keys = access_keys[i + 1 :]
                return [self.access(rest_keys, e) for e in d[k]]
            elif k.isdecimal():
                try:
                    d = d[int(k)]
                except IndexError:
                    return default
            else:
                try:
                    d = d[k]
                except KeyError:
                    return default
        return d


def normalize_json_pointer(ref):
    if "~" not in ref:
        return ref
    return ref.replace("~1", "/").replace("~0", "~")



import json
from json.decoder import WHITESPACE
from collections import deque, OrderedDict

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError  # for 3.4
from .rec import consume_rec


def load(stream, *, buffered=False):
    if buffered:
        return _load_buffered(stream)
    else:
        return _load_unbuffered(stream)


def _load_unbuffered(stream):
    buf = deque([], maxlen=100)
    first_err = None

    decoder = json.JSONDecoder(object_pairs_hook=OrderedDict)

    for line in stream:
        buf.append(line)
        try:
            body = "".join(buf)
            if len(body.strip()) > 0:
                idx = WHITESPACE.match(body).end()
                ob, end = decoder.raw_decode(body, idx)
                first_err = None
                yield ob
                buf.clear()
                buf.append(body[end:])
        except JSONDecodeError as e:
            if first_err is None:
                first_err = e

    if first_err is not None:
        raise first_err


def _load_buffered(stream):
    s = stream.read()
    size = len(s)
    decoder = json.JSONDecoder(object_pairs_hook=OrderedDict)

    end = 0
    while True:
        idx = WHITESPACE.match(s[end:]).end()
        i = end + idx
        if i >= size:
            break
        ob, end = decoder.raw_decode(s, i)
        yield ob


def dump(d, fp, *, squash_level=0, raw=False, extra_kwargs=None):
    opts = extra_kwargs or dict(sort_keys=False, ensure_ascii=False)

    def _dump(d):
        if raw and isinstance(d, str):
            print(d, file=fp)
        else:
            json.dump(d, fp=fp, **opts)
            print(file=fp)

    consume_rec(d, _dump, n=squash_level)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, "README.md")) as f:
        README = f.read()
    with open(os.path.join(here, "CHANGES.txt")) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ""

install_requires = ["magicalimport"]

docs_extras = []

yaml_extras = ["PyYAML"]

tests_require = []

testing_extras = tests_require + []

setup(
    name="jqfpy",
    version="0.5.2-1",
    description="jq for pythonista",
    long_description=README + "\n\n" + CHANGES,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="jq",
    author="podhmo",
    author_email="ababjam61+github@gmail.com",
    url="https://github.com/podhmo/jqfpy",
    packages=find_packages(exclude=["jqfpy.tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        "testing": testing_extras,
        "yaml": yaml_extras,
        "docs": docs_extras,
        "dev": ["black"],
    },
    tests_require=tests_require,
    test_suite="jqfpy.tests",
    entry_points="""
    [console_scripts]
    jqfpy=jqfpy.__main__:main
""",
)



import os
import sys
import contextlib
import argparse
import magicalimport
import jqfpy
from jqfpy import loading


def _describe_pycode(pycode, *, indent="", fp=sys.stderr):
    print(indent + pycode.replace("\n", "\n" + indent), file=fp)


@contextlib.contextmanager
def gentle_error_reporting(pycode, fp):
    try:
        yield
    except Exception as e:
        fp = sys.stderr
        print("\x1b[32m\x1b[1mcode:\x1b[0m", file=fp)
        print("----------------------------------------", file=fp)
        _describe_pycode(pycode, fp=fp, indent="")
        print("----------------------------------------", file=fp)
        print("", file=fp)
        print("\x1b[32m\x1b[1merror:\x1b[0m", file=fp)
        raise


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("code", nargs="?", default="get()")
    parser.add_argument("file", nargs="*", type=argparse.FileType("r"))
    parser.add_argument(
        "-i", "--input-format", choices=["json", "yaml", "ltsv"], default="json"
    )
    parser.add_argument(
        "-o", "--output-format", choices=["json", "yaml", "ltsv"], default="json"
    )

    parser.add_argument("-c", "--compact-output", action="store_true")
    parser.add_argument("-s", "--slurp", action="store_true")
    parser.add_argument("-S", "--sort-keys", action="store_true")
    parser.add_argument("-a", "--ascii-output", action="store_true")
    parser.add_argument("-r", "--raw-output", action="store_true")
    parser.add_argument(
        "--relative-path",
        action="store_true",
        help="when h.dumpfile(), iff accessing opend filename, treating as relative path",
    )
    parser.add_argument("--here", default=None, help="cwd for h.dumpfile()")

    parser.add_argument("--buffered", action="store_true", dest="buffered")
    parser.add_argument("--unbuffered", action="store_false", dest="buffered")
    parser.set_defaults(buffered=True)

    parser.add_argument("--squash", action="count", default=0)
    parser.add_argument("--show-code", action="store_true")
    parser.add_argument("--additionals")

    args = parser.parse_args()

    fnname = "_transform"
    if not args.file and os.path.exists(args.code):
        args.file.append(open(args.code))
        args.code = "get()"

    pycode = jqfpy.create_pycode(fnname, args.code)
    fp = sys.stdout

    if args.show_code:
        _describe_pycode(pycode, fp=fp, indent="")
        sys.exit(0)

    if args.file:
        files = args.file[:]
    else:
        files = [sys.stdin]

    additionals = None
    if args.additionals is not None:
        additionals = magicalimport.import_module(args.additionals)

    dump_extra_kwargs = dict(
        indent=None if args.compact_output else 2,
        sort_keys=args.sort_keys,
        ensure_ascii=args.ascii_output,
    )
    # xxx: chdir if here is not None
    ctx = jqfpy.create_context(here=args.here, extra_kwargs=dump_extra_kwargs)

    with gentle_error_reporting(pycode, fp):
        transform_fn = jqfpy.exec_pycode(fnname, pycode)

    def _load(streams, *, relative=args.relative_path):
        for stream in streams:
            if relative:
                filepath = loading.get_filepath_from_stream(stream)
                if filepath:
                    ctx.chdir(os.path.dirname(filepath))

            m = ctx.get_module(stream, format=args.input_format)
            for d in m.load(stream, buffered=args.buffered):
                yield d

    def _dump(d, *, i=0, fp=fp, raw=False):
        m = ctx.get_module(fp, format=args.output_format)
        if i > 0 and m.SEPARATOR:
            fp.write(m.SEPARATOR)
        m.dump(
            d,
            fp=fp,
            squash_level=args.squash,
            raw=raw or args.raw_output,
            extra_kwargs=ctx.dump_extra_kwargs,
        )
        if not args.buffered:
            fp.flush()

    if args.slurp:
        d = list(_load(files))
        with gentle_error_reporting(pycode, fp):
            r = jqfpy.transform(ctx, transform_fn, d, additionals=additionals)
        _dump(r, i=0)
    else:
        with gentle_error_reporting(pycode, fp):
            for i, d in enumerate(_load(files)):
                r = jqfpy.transform(ctx, transform_fn, d, additionals=additionals)
                _dump(r, i=i)
    fp.flush()


if __name__ == "__main__":
    main()



import os.path
import sys


class Repository:
    def __init__(self):
        self.mapping = {}
        self._cached = {}
        self.factories = {}

    def register(self, *exts):
        def _register(fn):
            self.factories[fn.__name__] = fn
            for ext in exts:
                self.mapping[ext] = fn.__name__
            return fn

        return _register

    def lookup(self, name):
        m = self._cached.get(name)
        if m is None:
            m = self._cached[name] = self.factories[name]()
        return m

    def lookup_by_extname(self, extname):
        return self.lookup(self.mapping[extname])


class Dispatcher:
    def __init__(self, repository):
        self.repository = repository

    def dispatch(self, stream, *, default_format="json"):
        return self._lookup(stream, default_format)

    def _lookup(self, stream, default):
        filename = get_filepath_from_stream(stream)
        ext = os.path.splitext(filename)[1]
        try:
            return self.repository.lookup_by_extname(ext)
        except KeyError:
            return self.repository.lookup(default)


_repo = Repository()


@_repo.register(".json", ".js")
def json():
    from . import _json as m

    m.SEPARATOR = None
    return m


@_repo.register(".yaml", ".yml")
def yaml():
    try:
        from . import _yaml as m

        m.SEPARATOR = "---\n"
    except ImportError:
        fp = sys.stderr
        print(
            "\x1b[33m\x1b[1myaml module is not found. please install via \n  pip install 'jqfpy[yaml]'\x1b[0m",
            file=fp,
        )
        sys.exit(1)
    return m


@_repo.register(".ltsv")
def ltsv():
    from . import _ltsv as m

    m.SEPARATOR = None
    return m


register = _repo.register
get_module = Dispatcher(_repo).dispatch


def get_filepath_from_stream(stream):
    return getattr(stream, "name", "")



from collections import OrderedDict
from .rec import consume_rec


def load(stream, *, buffered=False):
    for line in stream:
        yield OrderedDict(
            pair.split(":", 1) for pair in line.rstrip("\n\r").split("\t")
        )


def dump(d, fp, *, squash_level=0, raw=False, extra_kwargs=None):
    opts = extra_kwargs or dict(sort_keys=False, ensure_ascii=False)

    def _dump(d):
        if raw and isinstance(d, str):
            print(d, file=fp)
        else:
            if opts.get("sort_keys", False):
                items = sorted(d.items(), key=lambda pair: pair[0])
            else:
                items = d.items()
            print("\t".join("{}:{}".format(k, v) for k, v in items), file=fp)

    consume_rec(d, _dump, n=squash_level)
